package Apache::Log::Parser;
use 5.008005;
use strict;
use warnings;

our $VERSION = "0.01";



1;
__END__

=encoding utf-8

=head1 NAME

Apache::Log::Parser - It's new $module

=head1 SYNOPSIS

    use Apache::Log::Parser;

=head1 DESCRIPTION

Apache::Log::Parser is ...

=head1 LICENSE

Copyright (C) TAGOMORI Satoshi.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 AUTHOR

TAGOMORI Satoshi E<lt>tagomoris@gmail.comE<gt>

=cut

